"""Convenience to use colander with kerno."""

import colander as c

from kerno.state import MalbonaRezulto
from kerno.typing import DictStr

_ = str  # TODO add i18n


def validate_schema(
    schema: c.SchemaType,
    adict: DictStr,
    mal_title: str = _("Validation error"),
    mal_plain: str = _("The data do not pass server validation."),
) -> DictStr:
    """Conveniently validate a colander schema and return the clean dict.

    But if colander.Invalid is raised, put it inside a MalbonaRezulto.
    """
    try:
        return schema.deserialize(adict)
    except c.Invalid as e:
        malbona = MalbonaRezulto()
        malbona.add_toast(title=mal_title, plain=mal_plain)
        malbona.invalid = e.asdict()
        raise malbona


class InvalidToMalbona:
    """Context manager that wraps colander's Invalid in a MalbonaRezulto."""

    def __init__(
        self,
        title: str = _("Validation error"),
        plain: str = _("The data do not pass server validation."),
        html: str = "",
    ) -> None:  # noqa
        self.title = title
        self.plain = plain
        self.html = html

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        if isinstance(exc_val, c.Invalid):
            malbona = MalbonaRezulto()
            malbona.add_toast(
                title=self.title, plain=self.plain, html=self.html
            )
            malbona.invalid = exc_val.asdict()
            raise malbona


class NumLines:
    """Colander validator that checks the number of lines in text."""

    def __init__(
        self,
        min: int = 0,
        max: int = -1,
        min_err: str = "Not enough lines (minimum {min})",
        max_err: str = "Too many lines (maximum {max})",
    ) -> None:  # noqa
        self.min = min
        self.max = max
        self.min_err = min_err
        self.max_err = max_err

    def __call__(self, node, value) -> None:  # noqa
        num_lines = value.count("\n") + 1
        if self.min > 0 and num_lines < self.min:
            # err = _(self.min_err, mapping={"min": self.min})
            raise c.Invalid(node, self.min_err.format(min=self.min))
        if self.max > -1 and num_lines > self.max:
            # err = _(self.max_err, mapping={"max": self.max})
            raise c.Invalid(node, self.max_err.format(max=self.max))


class IsWeb:
    """Colander validator: ensure starts with "http://" or "https://"."""

    def __init__(self, msg: str = 'Must start with "http://" or "https://".'):
        self.msg = msg

    def __call__(self, node, value) -> None:  # noqa
        if not (value.startswith("http://") or value.startswith("https://")):
            raise c.Invalid(node, self.msg)



"""A complement to our email module, for those that use Pyramid."""

from bag.reify import reify
import premailer
from pyramid.renderers import render


class PyramidEmail:
    """Mixin that uses Pyramid to render the HTML template."""

    @reify
    def html(self):
        """Return the rich text version of the email message."""
        html = render(self.HTML_TEMPLATE, self.adict)
        return premailer.transform(html)  # Inline CSS styles



"""The Kerno class."""

from types import MappingProxyType  # which behaves like a FrozenDict
from typing import Any, Dict, Optional

from kerno.event import EventHub
from kerno.typing import DictStr


class Kerno:
    """Core of an application, integrating decoupled resources.

    The Kerno instance is used at runtime; at startup it is instantiated by
    the "Eko" configurator.
    """

    def __init__(
        self, settings: Dict[str, Dict], const: Optional[DictStr] = None
    ):
        """Construct. The ``settings`` are a dict of dicts."""
        self.settings = settings
        self.utilities: MappingProxyType[str, Any] = MappingProxyType({})
        self.const = const or {}  # The app should put global constants here
        self.events = EventHub()



"""A tiny event hub.

``kerno.events`` will be an instance of EventHub, however you can also
use the EventHub class in other circumstances if you want to.

First a library defines an event class to contain event data::

    class EventUserLoggedIn:
        def __init__(self, peto):
            self.peto = peto
            # ...and other data that user code might find important.

Elsewhere, user code will import the event class and write a handler::

    from example_library import EventUserLoggedIn

    def when_user_logs_in(event: EventUserLoggedIn):
        print(event.peto)

Finally, in setup code::

    kerno.events.subscribe(EventUserLoggedIn, when_user_logs_in)

Sometimes it is necessary to avoid leaks by unsubscribing::

    kerno.events.unsubscribe(EventUserLoggedIn, when_user_logs_in)

One of the advantages of having the event hub as a separate object is
that, instead of unsubscribing many handlers, sometimes you can just
throw away the hub, replacing its instance.

The library fires the event by doing::

    kerno.events.broadcast(EventUserLoggedIn(peto=peto))
"""

from typing import Callable, Dict, List


class EventHub:
    """A hub for events to be subscribed, fired and removed."""

    def __init__(self):  # noqa
        self._events: Dict[type, List[Callable]] = {}

    def subscribe(self, event_cls: type, function: Callable) -> Callable:
        """Subscribe a handler ``function`` to the ``event_cls``."""
        handlers: List[Callable] = self._events.setdefault(event_cls, [])
        if function in handlers:
            raise RuntimeError(
                f"This function is already subscribed to {event_cls}."
            )
        handlers.append(function)
        return function

    def unsubscribe(self, event_cls: type, function: Callable) -> bool:
        """Remove a function.  Return True if it really was subscribed."""
        handlers: List[Callable] = self._events.setdefault(event_cls, [])
        ret = function in handlers
        if ret:
            handlers.remove(function)
        return ret

    def broadcast(self, event):
        """Trigger/fire ``event`` -- execute its subscribers.

        The type of ``event`` must be an exact match: inheritance
        is not supported.
        """
        handlers: List[Callable] = self._events.setdefault(type(event), [])
        for fn in handlers:
            fn(event)



"""Integration between Kerno and the awesome Pyramid web framework."""

from functools import wraps
import inspect
from json import dumps
from typing import Callable

from bag.web.exceptions import Problem
from zope.interface import Interface

from kerno.kerno import Kerno
from kerno.state import MalbonaRezulto, Rezulto, to_dict
from kerno.typing import DictStr


def kerno_view(fn: Callable) -> Callable:
    """Decorate Pyramid views that call Kerno actions or operations.

    The view can return a Rezulto or RAISE a MalbonaRezulto. Then this
    decorator sets the status_int of the response (to 200 or 201) and
    converts it to a dictionary.
    """
    args = [str(name) for name in inspect.signature(fn).parameters.keys()]
    if len(args) == 1 and "self" == args[0]:  # view signature #1: self

        def get_request(args):
            return args[0].request

    elif len(args) == 1 and "request" == args[0]:  # view signature #2: request

        def get_request(args):
            return args[0]

    elif len(args) == 2 and "request" == args[1]:  # view signature #3:

        def get_request(args):  # context, request
            return args[1]

    else:
        raise RuntimeError(
            f"The kerno_view decorator found an unexpected signature in {fn}"
        )

    @wraps(fn)
    def wrapper(*args):
        try:
            rezulto = fn(*args)
        except Problem as e:
            adict = e.to_dict()
            raise MalbonaRezulto(
                # content_type='application/json',
                status_int=e.status_int,
                invalid=dumps(adict),
                title=e.error_title or "Server error",
                plain=e.error_msg,  # could be shown to end users
                debug=e.error_debug,  # not displayed to end users
            )
        if rezulto is None:
            raise RuntimeError(
                "Error: None returned by view {}()".format(fn.__qualname__)
            )
        elif isinstance(rezulto, Rezulto):
            request = get_request(args)
            request.response.status_int = rezulto.status_int
            return to_dict(obj=rezulto)
        else:
            return rezulto

    return wrapper


def malbona_view(context, request) -> DictStr:
    """Pyramid view handler that returns a MalbonaRezulto as a dictionary."""
    request.response.status_int = context.status_int
    return to_dict(context)


class IKerno(Interface):
    """Marker to register and retrieve a Kerno instance in a Pyramid app."""


def includeme(config) -> None:
    r"""Integrate kerno with Pyramid.

    - Make ``request.kerno`` available.
    - Make ``request.repo`` available.
    - Also register an ``IKerno`` interface so one can retrieve the kerno
      instance from the Pyramid registry with
      ``kerno = registry.queryUtility(IKerno, default=None)``.
    - Add our views to a Pyramid app so it will display our exceptions.

    Usage example::

        from kerno.web.pyramid import IKerno, Kerno

        def init_kerno(config_path: str) -> Kerno:
            \"\"\"Return initialized kerno, separate from web frameworks.\"\"\"
            from kerno.start import Eko
            eko = Eko.from_ini(config_path)
            # ...more kerno initialization code here, then...
            return eko.kerno

        def main(global_config, **settings):
            \"\"\"Return a Pyramid WSGI application.\"\"\"
            from pyramid.config import Configurator
            with Configurator(settings=settings) as config:
                kerno = init_kerno(global_config['__file__'])
                config.registry.registerUtility(kerno, IKerno)
                config.include('kerno.web.pyramid')
                # ...more Pyramid configuration code here, then at the end...
            return config.make_wsgi_app()

    For the HTML version, if you'd like to change our template to your own,
    just override the Pyramid view configuration. Example::

        config.add_view(
            context=MalbonaRezulto, accept='text/html', view=malbona_view,
            renderer='yourapp:templates/malbona.jinja2')
    """
    kerno = config.registry.queryUtility(IKerno, default=None)
    if not isinstance(kerno, Kerno):
        raise RuntimeError(
            "A Kerno instance must be registered with Pyramid "
            "before including 'kerno.web.pyramid'."
        )

    config.add_request_method(  # request.kerno is computed once per request
        lambda request: request.registry.getUtility(IKerno),
        "kerno",
        reify=True,
    )

    config.add_request_method(  # request.repo is computed once per request
        lambda request: request.kerno.new_repo(), "repo", reify=True
    )

    config.registry.registerUtility(kerno, IKerno)

    config.add_view(
        context=MalbonaRezulto,
        accept="text/html",
        view=malbona_view,
        renderer="kerno:web/malbona.jinja2",
    )
    config.add_view(
        context=MalbonaRezulto,
        renderer="json",
        view=malbona_view,
        accept="application/json",
    )
    config.add_view(
        context=MalbonaRezulto, renderer="json", view=malbona_view, xhr=True
    )



from dataclasses import dataclass
from typing import Any, Optional, Type, TypeVar

from kerno.kerno import Kerno
from kerno.repository.sqlalchemy import BaseSQLAlchemyRepository
from kerno.typing import DictStr

# A generic variable that can be Peto or any subclass
APeto = TypeVar("APeto", bound="Peto")


@dataclass
class Peto:
    """An object to store incoming state.

    A Peto is typically instantiated in the controller layer and then
    passed around (as the main context object) to actions, event handlers,
    schemas, helpers etc.  This saves you from lots of boilerplate,
    removes global variables and makes unit testing easier.

    Look at the code for this class as an example.  It is recommended that
    you subclass with more concrete typing annotations.
    For instance, use static typing for the ``user`` variable.
    """

    kerno: Kerno  # the global application object
    repo: BaseSQLAlchemyRepository  # data access layer
    user: Any = None  # the current user requesting an operation
    raw: Optional[DictStr] = None  # dictionary of operation-specific data

    @classmethod
    def from_pyramid(cls: Type[APeto], request, json=False) -> APeto:
        """Integration with the *pyramid* web framework.

        This comes free of any pyramid imports.

        Typical usage is, inside your pyramid view, you do::

            peto = Peto.from_pyramid(request, json=True)
            # Call your action/service layer, which returns a Rezulto instance:
            rezulto = create_user(peto)
            return rezulto

        Use the ``json`` flag to obtain the payload from the pyramid request.

        Another example, very brief::

            @kerno_view
            def get_audit_data(request):
                "\""Return audit data to be displayed to a superuser."\""
                return get_audits(
                    peto=Peto.from_pyramid(request), **request.json_body)
        """
        return cls(
            kerno=request.kerno,
            repo=request.repo,
            user=request.user,
            raw=request.json_body if json else None,
        )



"""A better way to send entities through the wire.

What problems does this solve?

1. Python model classes do not need to implement ``as_dict()``. Instead, you
implement a variation of a ``jsonright()`` function in a separate step,
which could even be in a separate package, totally decoupled. This does
`single dispatch`_ -- please learn about this.

.. _`single dispatch`: https://docs.python.org/3/library/functools.html#functools.singledispatch

2. When we send objects through the wire, different detail levels are
commonly needed -- for instance, first we only want the names of users to be
shown in a list, and then in a separate step, we might want all the details
of a single user. Your ``jsonright()`` implementation receives a ``features``
argument which can be a set of strings (or a sequence of any type)
specifying the level of detail. This keeps the API uniform.

3. This module also offers tools to help you write those
``jsonright()`` implementations -- especially ``entity2dict()``.

4. Sending common JSON through the wire has a problem: the repetition of keys.
For example, if you output a list of users, the keys of each object, such as
"id", "name", "email" etc. are repeated for each instance in the JSON.
Here we save bandwidth by encoding the data differently and then
reassembling it in the client.

5. The ``jsonright()`` interface has a ``peto`` argument, therefore when
implementing it you have access to the context, including application
configuration, repository (data access layer), current user etc.

How are data organized to save bandwidth?
=========================================

jsonright avoids repeating key names when outputting sequences of entities
in JSON. It pivots the table so it grows to the right::

    [
        ["id", 1, 2],
        ["email", "ex@am.pl", "sagan@nasa.gov"],
    ]

Of course objects are easily reassembled in Javascript.  The function that
does that is in the file kerno.js.

Usage
=====

Please see a usage example in
`our tests <https://github.com/nandoflorestan/kerno/blob/master/tests/test_web_jsonright.py>`_.
"""

from datetime import date, datetime
from decimal import Decimal
from functools import singledispatch
from typing import Any, Iterable, Sequence

from bag import first
from kerno.peto import Peto
from kerno.typing import DictStr


def keys_from(obj: Any) -> Iterable[str]:
    """Return the names of the instance variables of ``obj``."""
    return vars(obj).keys()


def only_relevant(keys: Iterable[str]) -> Iterable[str]:
    """Ignore strings that start in dunder ("__") or in "_sa_".

    These usually keep SQLAlchemy state.
    """
    return filter(
        lambda key: not key.startswith("__") and not key.startswith("_sa_"),
        keys,
    )


def excluding(
    blacklist: Sequence[str], keys: Iterable[str]
) -> Iterable[str]:  # noqa
    return filter(lambda k: k not in blacklist, keys)


def entity2dict(
    obj: Any,
    keys: Iterable[str] = (),
) -> DictStr:
    """Dump certain instance variables of ``obj`` into a dictionary.

    This function is reusable.

    If you do not provide any ``keys``, a sensible default is used.
    """
    kk = keys or excluding(("password",), only_relevant(keys_from(obj)))
    return {key: getattr(obj, key) for key in kk}


@singledispatch
def jsonright(obj: Any, peto: Peto, features=(), **kw) -> Any:
    """Overloadable function to encode entities for sending through the wire.

    You can register your own implementations which get called depending
    on the type of *obj*.
    """
    raise NotImplementedError(
        f"No implementation of jsonright() registered for type {type(obj)}"
    )


@jsonright.register(str)
@jsonright.register(int)
@jsonright.register(float)
@jsonright.register(bool)
@jsonright.register(type(None))
def _a(obj, peto: Peto, features=(), **kw) -> Any:
    return obj


@jsonright.register(bytes)
def _b(obj, peto: Peto, features=(), **kw) -> Any:
    return obj.decode(kw.get("encoding", "utf-8"))


@jsonright.register(Decimal)
def _c(obj, peto: Peto, features=(), **kw) -> float:
    return float(str(obj))


@jsonright.register(datetime)
@jsonright.register(date)
def _d(obj, peto: Peto, features=(), **kw) -> str:
    return obj.isoformat()


@jsonright.register(dict)
def _e(obj, peto: Peto, features=(), **kw) -> DictStr:
    return {
        str(key): jsonright(val, peto, features, **kw)
        for (key, val) in obj.items()
    }


primitive_types = (
    str,
    bytes,
    int,
    float,
    bool,
    type(None),
    Decimal,
    date,
    datetime,
    dict,
)


@jsonright.register(list)
@jsonright.register(tuple)
@jsonright.register(set)
@jsonright.register(frozenset)
def _s(obj, peto: Peto, features=(), **kw) -> Sequence:
    if len(obj) == 0:
        return []
    first_item = first(obj)
    if isinstance(first_item, primitive_types):
        return [jsonright(item, peto, features, **kw) for item in obj]
    # Below this line we assume we are dealing with a sequence of entities.
    # In this case we pivot data in order to save bandwidth.
    first_dict = jsonright(first_item, peto, features, **kw)
    ret = [[key] for key in first_dict]
    for entity in obj:
        adict = jsonright(entity, peto, features, **kw)
        for alist in ret:
            alist.append(jsonright(adict[alist[0]], peto, features, **kw))
    return ret



"""Eko is the class used to start a Kerno application.

It works much like initial configuration of a Pyramid app.
"""

from configparser import NoSectionError
from types import ModuleType
from typing import Callable, Iterable, List, Union

from bag.settings import read_ini_files, resolve

from kerno.kerno import Kerno
from kerno.typing import DictStr
from kerno.utility_registry import ConfigurationError, UtilityRegistryBuilder

ResourceType = Union[str, ModuleType, Callable]


class Eko:
    r"""At startup builds the Kerno instance for the app.

    In the Esperanto language, Eko is a noun that means "a start".
    And "eki" is a verb that means "to start".

    The Eko class can build an application core from multiple Python modules,
    much like the Configurator in Pyramid.

    If the same module is included twice, Eko will raise
    ``kerno.start.ConfigurationError``, helping you ensure
    your startup code stays clean.

    Usage example::

        from kerno.start import Eko
        eko = Eko.from_ini('main_config.ini', 'production.ini')
        eko.settings  # lets you access the merged settings from INI files
        eko.kerno  # is the Kerno instance that will be the core of the app

        eko.utilities.register('mailer', 'some.package:mailer_function')
        # ...and later you can retrieve *mailer_function* by doing:
        # kerno.utilities['mailer']

        # This finds the function some.extension.package:eki() and runs it,
        # passing it the eko configurator object:
        eko.include('some.extension.package')

        # For instance, here is how you can build a repository from 2 classes:
        eko.include('kerno.repository')  # adds add_repository_mixin() to eko
        eko.add_repository_mixin(
            'kerno.repository.sqlalchemy.BaseSQLAlchemyRepository')
        eko.add_repository_mixin('my.package:MyRepoMixinClass')
        # After this, the first time you access ``kerno.Repository``,
        # the Repository class is assembled from the mixin classes,
        # and stored for usage as kerno.Repository.

    If you need to debug the order in which modules got included, you can::

        print(*eko._included_modules, sep='\n')
    """

    @classmethod
    def from_ini(cls, *config_files, encoding: str = "utf-8"):
        """Return an instance after reading some INI file(s)."""
        return cls(settings=read_ini_files(*config_files, encoding=encoding))

    def __init__(self, settings: DictStr = {}):  # noqa
        if settings and not hasattr(settings, "__getitem__"):
            raise TypeError(
                "The *settings* argument must be dict-like. "
                "Received: {}".format(type(settings))
            )

        self._included_modules: List[ModuleType] = []
        self.kerno = Kerno(settings)
        self.utilities = UtilityRegistryBuilder(kerno=self.kerno)

        try:
            main_config_section = settings["kerno"]
        except (NoSectionError, KeyError):
            return
        for extension in main_config_section.get("includes", []):
            self.include(extension)

    def include(self, resource: ResourceType, throw: bool = True) -> None:
        """Execute a configuration callable for imperative extension.

        If the argument ``throw`` is True (which is the default)
        and ``resource`` does not have an ``eki()`` function,
        raise ConfigurationError.  Otherwise, ignore the current module
        and return without an error.
        """
        # TODO log.debug('Including {}'.format(module))
        obj = resolve(resource) if isinstance(resource, str) else resource
        if obj in self._included_modules:
            raise ConfigurationError(f"{obj} has already been included!")
        else:
            self._included_modules.append(obj)

        if isinstance(obj, ModuleType):
            try:
                fn = getattr(obj, "eki")
            except AttributeError:
                if throw:
                    raise ConfigurationError(
                        "The module {} has no function called 'eki'.".format(
                            obj.__name__
                        )
                    )
                else:
                    return
        else:
            fn = obj
        fn(self)

    def include_many(
        self, resources: Iterable[ResourceType], throw: bool = True
    ) -> None:
        """Initialize multiple app modules."""
        for resource in resources:
            self.include(resource, throw=throw)



"""A complement to our email module, for those who use pyramid_mailer."""

from typing import Dict

from pyramid_mailer.message import Message

from kerno.email import EmailMessageBase
from kerno.typing import DictStr


class EmailMessage(EmailMessageBase):
    """Abstract base class with methods that depend on pyramid_mailer."""

    def __init__(self, mailer_settings: Dict[str, str], **kw):  # noqa
        super().__init__(**kw)
        self.mailer_settings = mailer_settings

    def to_message_args(self) -> DictStr:
        """Convert this into a dict to instantiate a pyramid_mailer Message.

        This is useful because Celery likes to use JSON serialization.
        """
        adict = self.to_dict()
        env = adict['envelope']
        args = {
            'subject': adict['subject'],
            'html': adict['html'],
            'body': adict['plain'],
            'recipients': env['recipients'],
            'sender': env['sender'] or self.mailer_settings["default_sender"],
            'cc': env['cc'],
            'bcc': env['bcc'],
        }
        if env['reply_to']:
            args['extra_headers'] = ('Reply-To', env['reply_to']),
        return args

    def to_message(self) -> Message:
        """Convert this instance into a pyramid_mailer Message."""
        margs = self.to_message_args()
        return Message(**margs)



"""An abstract base for class-based actions."""

from abc import ABCMeta

from kerno.peto import Peto


class Action(metaclass=ABCMeta):
    """Abstract base for class-based actions.

    Subclasses must implement __call__() and, if happy,
    return a Rezulto instance::

        from kerno.action import Action
        from kerno.state import Rezulto

        class MyAction(Action):
            def __call__(self, *a, **kw) -> Rezulto:
                ...
    """

    def __init__(self, peto: Peto):  # noqa
        self.peto = peto

    @property
    def kerno(self):
        """Return the global application object."""
        return self.peto.kerno

    @property
    def repo(self):
        """Return a repository instance for the current request."""
        return self.peto.repo

    @property
    def user(self):
        """Return the current user or None."""
        return self.peto.user



"""Repositories are persistence strategies.

The repository pattern describes a storage component that can be easily
swapped through dependency injection.

The purposes of a repository are:

1. To give a name to each query so the code becomes easier to read
2. To isolate data access code
3. Such that you can easily write unit tests (not integration tests) for the
   action / business layer. Because it is much, much easier to fake a
   repository than it is to fake SQLAlchemy and other ORMs.

Tests become easier to write because a repository is
dependency-injected into the action/service layer.

Extracting a repository from existing code does reveal repeated queries -- and
repetition is the root of all evil in software creation.

Notably absent from the above list of purposes is the dream of
easily swapping data access technology. The repository pattern does not
make the dream come true because persistence frameworks have their own
ideas about the models, too. For instance:

- ZODB is for when you traverse a lot; SQLALchemy is for when you search
  and query a lot. Usually it doesn't make sense to keep swapping
  between these two because they have different use cases.
- In SQLALchemy you model many-to-many relationships through an association
  table; in ZODB you usually do this in a very different way. Repositories
  for each tech would need to capture this.
- ZODB is very Pythonic, you include a lot of code in constructors.
  But SQLAlchemy is best used without constructors in the model classes.

In the end these differences make it very hard to develop software
that is so abstract that it deals with both techs well.

If in addition to different repositories you need to develop different models
for each tech, then there is no real code reuse, is there?

So the main purpose of a repository is to isolate IO so you can write
real unit tests for the pure part of the code.

I also recommend against building generic repositories. See
http://ben-morris.com/why-the-generic-repository-is-just-a-lazy-anti-pattern
"""

from typing import Iterable

from bag.settings import resolve

from kerno.start import Eko


def compose_class(name: str, mixins: Iterable) -> type:
    """Return a class called ``name``, made of the bases ``mixins``."""
    bases = (resolve(mixin) if isinstance(mixin, str) else mixin
             for mixin in mixins)
    return type(name, tuple(bases), {})


def eki(eko: Eko) -> None:
    """Make repository functionality available.

    - *eko* gets the ``add_repository_mixin(cls)`` method which is used to
      gradually build the ``kerno.Repository`` class from modular classes.
    - *kerno* gets a ``new_repo()`` method which instantiates a Repository.
    """
    eko._repository_mixins = []  # type: ignore

    def add_repository_mixin(mixin):
        """Store one of the mixin classes to form the final repository."""
        assert isinstance(mixin, (str, type))
        eko._repository_mixins.append(mixin)
        # When kerno.Repository is first accessed, the class will be assembled
        # (only once) and will stay as a variable of the kerno instance:
        eko.kerno.Repository = compose_class(
            name='Repository', mixins=eko._repository_mixins)
    eko.add_repository_mixin = add_repository_mixin  # type: ignore

    eko.kerno.new_repo = (  # type: ignore
        lambda: eko.kerno.Repository(kerno=eko.kerno))  # type: ignore



"""A setup so you can have many, powerful to_dict() implementations.

Our main objective is to **remove any JSON concerns from model classes**.
These should NOT contain ``as_dict()`` implementations because the system
may need multiple converters for the same model -- for instance,
with varying fields, verbosities, levels of detail,
optional inclusion of related entities etc.

**It is bad if AVerySpecificModel.as_dict() needs a different signature**
to accept a parameter that all the other as_dict() methods do not need.

Here is the solution:

By using `Reg <http://reg.readthedocs.io/en/latest/>`_ we now have a
``to_dict(obj, flavor='', **kw)`` function that dispatches
on the type of ``obj`` and ``flavor``. This allows you to register
more than one implementation (with a "flavor" name) for each of your
model classes.

In order to understand this, you need to know what Reg does:
multiple dispatch.

Our to_dict() function also has a default implementation, so it can be used
directly.

This way you can create a situation in which::

    to_dict(Address, flavor="")  # calls one implementation, and
    to_dict(Person, flavor="")  # calls another implementation, and
    to_dict(Person, flavor="table")   # calls yet another implementation.

If the user code omits the *flavor* argument, the default one -- whose
value is an empty string -- gets used.

And because our ``to_dict()`` accepts keyword arguments, you can create
a very powerful version of it, that can take such arguments as
a repository, a user object, the current date etc.

Please see a usage example in
`our tests <https://github.com/nandoflorestan/kerno/blob/master/tests/test_web_to_dict.py>`_.
"""

from collections import OrderedDict
from datetime import date, datetime
from decimal import Decimal
from typing import Any, Iterable, Sequence
from warnings import warn

import reg

warn(
    "to_dict will be removed in a future version of kerno. "
    "Use jsonright instead.",
    DeprecationWarning,
)


def keys_from(obj: Any) -> Iterable:
    """Return the names of the instance variables of ``obj``."""
    return obj.__dict__.keys()


def only_relevant(keys: Iterable[str]) -> Iterable[str]:
    """Ignore strings that start in dunder ("__") or in "_sa_".

    These are usually keeping SQLAlchemy state.
    """
    return filter(
        lambda key: not key.startswith("__") and not key.startswith("_sa_"),
        keys,
    )


def excluding(blacklist: Sequence, keys: Iterable) -> Iterable:  # noqa
    return filter(lambda k: k not in blacklist, keys)


def reuse_dict(
    obj: Any,
    keys: Iterable = (),
    for_json: bool = True,
    sort: bool = True,
    **kw,
) -> OrderedDict:
    """Dump the instance variables of ``obj`` into an OrderedDict.

    This function is reusable and free of Reg dispatch.

    If the ``for_json`` flag is True, convert certain types.

    If the ``sort`` flag is True, sort the OrderedDict.
    """
    amap: OrderedDict[str, Any] = OrderedDict()
    kk = keys or excluding(("password",), only_relevant(keys_from(obj)))
    if sort:
        kk = sorted(kk)
    for key in kk:
        val = getattr(obj, key)
        # TODO Probably should treat Python types elsewhere!
        if for_json and (isinstance(val, datetime) or isinstance(val, date)):
            amap[key] = val.isoformat()
        elif for_json and isinstance(val, Decimal):
            amap[key] = float(str(val))
        elif for_json and not isinstance(
            val, (str, int, float, list, dict, bool, type(None))
        ):
            continue
        else:
            amap[key] = val
    return amap


@reg.dispatch(  # Dispatch on type of *obj* and value of *flavor*.
    reg.match_instance("obj"),
    reg.match_key("flavor", lambda obj, flavor, **kw: flavor),
)
# Cannot type-annotate this function, Reg 0.11 does not support it
def to_dict(obj, flavor="", **kw):
    """Overloadable version of our function ``reuse_dict``.

    You can register your own implementations depending on *obj* and *flavor*.
    """
    return reuse_dict(obj, **kw)


"""
Ideas
-----

- Use a JSON serializer without tree traversal, but support Python data types such as datetime
- convert datetime to {"_T": "Date", "val": "..."}... NO. Convert to string and the client must know that such and such attributes must be converted to Date. Maybe provide the mapping (or even JS code) as another string?
- provide SQLAlchemy converter based on mine -- No, because https://pypi.python.org/pypi/colander_jsonschema/
- Convert trees of objects
- optionally provide "_typ" in output... but better as a property of a collection?

Other JSON related projects
---------------------------

jsonplus has "exact" and "compat" serializers:
https://pypi.python.org/pypi/jsonplus/0.8.0

morejson converts to and from stdlib types:
https://pypi.python.org/pypi/morejson/1.1.5
https://pypi.python.org/pypi/jsontyping/1.0.3

https://github.com/atsuoishimoto/emitjson

https://pypi.python.org/pypi/hgijson/3.1.0

Annoying declarative API, but achieves a lot:
http://hgi-json.readthedocs.io/en/latest/functionality/

https://pypi.python.org/pypi/json_tricks/3.11.3


Outside my scope and purpose, but probably great
------------------------------------------------

jsonpickle does arbitrary trees (which we don't really want).
https://pypi.python.org/pypi/jsonpickle/

https://pypi.python.org/pypi/json-model/1.0.1
https://pypi.python.org/pypi/json-delta/2.0
https://pypi.python.org/pypi/json-merger/0.4.0
"""



"""Helpers and sane defaults for usage of the Pydantic library."""

from typing import TYPE_CHECKING

from pydantic import BaseModel, conint, constr

if TYPE_CHECKING:
    ReqStr = str
    EmptyStr = str
    ZeroOrMore = int
else:
    ReqStr = constr(min_length=1, strip_whitespace=True, strict=True)
    EmptyStr = constr(min_length=0, strip_whitespace=True, strict=True)
    ZeroOrMore = conint(gt=-1)


class Pydantic(BaseModel):
    """Base class for our validation models."""

    class Config:
        """Controls the behaviour of pydantic."""

        # https://pydantic-docs.helpmanual.io/usage/model_config/
        anystr_strip_whitespace = True



"""A registry to store various utilities in an application."""

from configparser import NoSectionError
from types import MappingProxyType  # behaves like a FrozenDict / frozendict
from typing import Any

from bag.settings import resolve

from kerno.kerno import Kerno
from kerno.typing import DictStr


class ConfigurationError(Exception):
    """Represents an error during application startup."""


class UtilityRegistryBuilder:
    """Gradually builds Kerno's utility registry, which is immutable."""

    def __init__(self, kerno: Kerno):
        """Read the config section "kerno utilites"; register each utility."""
        self.kerno = kerno
        self._utilities: DictStr = {}
        self.kerno.utilities = MappingProxyType(self._utilities)

        try:
            section = kerno.settings["kerno utilities"]
        except (NoSectionError, KeyError):
            return
        for name, utility in section.items():
            self.register(name, utility)

    def register(self, name: str, utility: Any) -> Any:
        """Register ``utility`` under ``name`` at startup for later use."""
        obj = resolve(utility)
        # print('Registering ', name, utility, obj)
        self._utilities[name] = obj
        return obj

    def set_default(self, name: str, utility: Any) -> Any:
        """Register ``utility`` as ``name`` only if name not yet registered."""
        if self._utilities.get(name) is None:
            return self.register(name, utility)
        else:
            return None

    def ensure(self, name: str, component: str = "The application") -> None:
        """Raise if no utility has been registered under ``name``."""
        if self._utilities.get(name) is None:
            raise ConfigurationError(
                f'{component} needs a utility called "{name}", '
                "which has not been registered."
            )



"""A registry for ways to convert an UIMessage to HTML code."""

from html import escape
import reg
from kerno.state import UIMessage
from kerno.web.to_dict import to_dict


# Dispatch depending on the value of *flavor*.
@reg.dispatch(reg.match_key("flavor", lambda flavor, msg, close: flavor))
# Cannot type-annotate this function, Reg 0.11 does not support it:
def msg_to_html(flavor, msg, close=True):
    """Render ``msg`` (an UIMessage instance) as HTML markup.

    Example usage in a Mako template::

        <div class="flash-messages">
            % for msg in request.session.pop_flash():
                ${msg_to_html(flavor='bootstrap3', msg=msg)|n}
            % endfor
        </div>

    Example usage in a Jinja2 template::

        {%- block flash -%}
            <div class="flash-messages">
                {%- for msg in request.session.pop_flash() -%}
                    {$ msg_to_html(flavor='bootstrap3', msg=msg) | safe $}
                {%- endfor -%}
            </div>
        {%- endblock -%}
    """
    raise NotImplementedError()


@msg_to_html.register(flavor="bootstrap3")
# Cannot type-annotate this function, Reg 0.11 does not support it:
def msg_to_bootstrap3(flavor, msg, close=True):
    """Render the UIMessage ``msg`` as bootstrap 3 compatible HTML.

    If using Pyramid you can store UIMessage instances as flash messages
    in the user's session, then use this to render them as bootstrap alerts.
    """
    return (
        '<div class="alert alert-{level}{cls} fade in">{close}'
        "{body}</div>\n".format(
            level=escape(msg.level),
            cls=" alert-block" if msg.html else "",
            close='<button type="button" class="close" data-dismiss="alert" '
            'aria-label="Close"><span aria-hidden="true">×</span>'
            "</button>"
            if close
            else "",
            body=msg.html or escape(msg.plain),
        )
    )


def includeme(config):
    """Make ``msg_to_html()`` available to templates in Pyramid.

    Also add the ``add_flash`` request method.
    """

    def before_rendering_template(event):
        event["msg_to_html"] = msg_to_html
        event["flash_msgs_as_dicts"] = lambda: [
            to_dict(obj=f) for f in event["request"].session.pop_flash()
        ]

    from pyramid.events import BeforeRender

    config.add_subscriber(before_rendering_template, BeforeRender)

    config.add_request_method(add_flash, "add_flash")


def add_flash(request, allow_duplicate: bool = False, **kw) -> UIMessage:
    """Add a flash message to the user's session. For convenience."""
    msg = UIMessage(**kw)
    request.session.flash(msg, allow_duplicate=allow_duplicate)
    return msg



"""A base class for SQLAlchemy-based repositories."""

from typing import Any, Generic, Iterable, List, Optional

from kerno.kerno import Kerno
from kerno.typing import DictStr, Entity


class BaseSQLAlchemyRepository:
    """Base class for a SQLAlchemy-based repository."""

    SAS = 'session factory'  # name of the SQLAlchemy session utility

    def __init__(self, kerno: Kerno, session_factory: Any = None):
        """Construct a SQLAlchemy repository instance to serve ONE request.

        - ``kerno`` is the Kerno instance for the current application.
        - ``session_factory`` is a function that returns a
          SQLAlchemy session to be used in this request -- scoped or not.
          If not provided as an argument, get it from the
          kerno utility registry (under the "session factory" name.
        """
        self.kerno = kerno
        self.sas = self.new_sas(
            session_factory or kerno.utilities[self.SAS])
        assert self.sas

    def new_sas(self, session_factory):
        """Obtain a new SQLAlchemy session instance."""
        assert session_factory is not None
        is_scoped_session = hasattr(session_factory, 'query')
        # Because we don't want to depend on SQLAlchemy:
        if callable(session_factory) and not is_scoped_session:
            return session_factory()
        else:
            return session_factory

    def add(self, entity: Entity) -> Entity:
        """Add an object to the SQLAlchemy session, then return it."""
        self.sas.add(entity)
        return entity

    def delete(self, entity: Entity) -> None:
        """Delete an ``entity`` from the database."""
        self.sas.delete(entity)

    def flush(self) -> None:
        """Obtain IDs on new objects and update state on the database.

        Without committing the transaction.
        """
        self.sas.flush()

    def _get_or_add(self, cls: type, **filters):
        """Retrieve or add an entity with ``filters``; return the entity.

        The entity will have a transient ``_is_new`` flag telling you whether
        it already existed.

        This is a helper for the implementation of repository methods and
        should not be used elsewhere.
        """
        entity = self.sas.query(cls).filter_by(**filters).first()
        if entity is None:
            entity = cls(**filters)
            self.add(entity=entity)
            is_new = True
        else:
            is_new = False
        assert not hasattr(entity, '_is_new')
        entity._is_new = is_new
        return entity

    def _update_or_add(self, cls: type, props: DictStr = {}, **filters):
        """Load and modify entity if it exists, else create one.

        First obtain either an existing object or a new one, based on
        ``filters``. Then apply ``props`` and return the entity.

        The entity will have a transient ``_is_new`` flag telling you whether
        it already existed.

        This is a helper for the implementation of repository methods and
        should not be used elsewhere.
        """
        assert '_is_new' not in props
        entity = self._get_or_add(cls, **filters)
        for key, val in props.items():
            setattr(entity, key, val)
        return entity


class Query(Iterable, Generic[Entity]):
    """Typing stub for a returned SQLAlchemy query.

    This is purposely very incomplete.  It is intended to be used as return
    value in repository methods, such that user code can use, but not change,
    the returned query, which is what we like to do in this architecture.

    If you want a more complete implementation, try
    https://github.com/dropbox/sqlalchemy-stubs

    Their query stub is at
    https://github.com/dropbox/sqlalchemy-stubs/blob/master/sqlalchemy-stubs/orm/query.pyi
    """

    def all(self) -> List[Entity]: ...  # noqa
    def count(self) -> int: ...  # noqa
    # def exists(self): ...  # noqa
    def first(self) -> Optional[Entity]: ...  # noqa
    def get(self, ident) -> Optional[Entity]: ...  # noqa
    def one(self) -> Entity: ...  # noqa
    # def slice(self, start: int, stop: Optional[int]): ...  # noqa
    def yield_per(self, count: int) -> List[Entity]: ...  # noqa



"""Classes that store outgoing state.

These are typically returned by *action* layer code to controllers.
"""

from abc import ABCMeta
from collections import OrderedDict
from typing import List, Optional

from kerno.typing import DictStr
from kerno.web.to_dict import to_dict, reuse_dict


class UIMessage:
    """Represents a message to be displayed to the user in the UI."""

    LEVELS = ["danger", "warning", "info", "success"]
    # The default to_dict() works fine for this class.

    def __init__(
        self,
        level: str = "danger",
        title: str = "",
        plain: str = "",
        html: str = "",
    ) -> None:
        """Constructor.

        ``level`` must be one of ("danger", "warning", "info", "success").
        """
        args_are_valid = (plain and not html) or (html and not plain)
        assert args_are_valid
        if level == "error":
            level = "danger"
        assert level in self.LEVELS, (
            'Unknown message level: "{0}". '
            "Possible levels are {1}".format(level, self.LEVELS)
        )
        self.level = level
        self.title = title
        self.plain = plain
        self.html = html

    def __repr__(self) -> str:
        return '<{} "{}">'.format(self.__class__.__name__, self.title)


class UICommand:
    """Represents a message telling the UI to do something."""

    __slots__ = ("name", "payload")

    def __init__(self, name: str, payload: DictStr) -> None:
        """Construct a message to the UI.

        ``name`` is the name of a command to be performed in the UI.
        ``payload`` is the data to run the command with.
        """
        self.name = name
        self.payload = payload

    def __repr__(self):
        return '<UICommand "{}">'.format(self.name)


@to_dict.register(obj=UICommand, flavor="")
def uicommand_to_dict(obj, flavor="", **kw):
    """Convert to dict a UICommand instance."""
    return OrderedDict((("name", obj.name), ("payload", obj.payload)))


class Returnable(metaclass=ABCMeta):
    """Base class for Rezulto and for MalbonaRezulto.

    Returnable is the base class for what Actions should return. It contains:

    - messages: Grave UI messages.
    - toasts: UI messages that disappear automatically after a while.
    - commands: messages to the UI, e. g., add an entity to your models
    - debug: A dict with information that is not displayed to the end user.
    - redirect: URL or screen to redirect to.

    Subclasses overload the ``status_int`` and ``level`` static variables.
    """

    level = "danger"
    status_int = 500  # HTTP response code indicating server bug/failure

    def __init__(
        self,
        commands: List[UICommand] = None,
        debug: DictStr = None,
        redirect: str = "",
        **kw,
    ):  # noqa
        self.messages: List[UIMessage] = []
        self.toasts: List[UIMessage] = []
        self.commands = commands or []
        self.debug = debug or {}
        self.redirect = redirect
        for k, v in kw.items():
            setattr(self, k, v)

    def __repr__(self) -> str:
        return "<{} status: {}>".format(
            self.__class__.__name__, self.status_int
        )

    def add_message(self, level: str = "", **kw) -> UIMessage:
        """Add to the grave messages to be displayed to the user on the UI."""
        msg = UIMessage(level=level or self.level, **kw)
        self.messages.append(msg)
        return msg

    def add_toast(self, level: str = "", **kw) -> UIMessage:
        """Add to the quick messages to be displayed to the user on the UI."""
        msg = UIMessage(level=level or self.level, **kw)
        self.toasts.append(msg)
        return msg

    def add_command(self, **kw) -> UICommand:
        """Add to the commands for the UI to perform."""
        cmd = UICommand(**kw)
        self.commands.append(cmd)
        return cmd


@to_dict.register(obj=Returnable, flavor="")
def returnable_to_dict(obj, flavor="", **kw):
    """Convert instance to a dictionary, usually for JSON output."""
    amap = reuse_dict(
        obj=obj,
        keys=kw.get("keys", ("level", "status_int", "debug", "redirect")),
        sort=False,
    )
    amap["messages"] = [reuse_dict(obj=msg) for msg in obj.messages]
    amap["toasts"] = [reuse_dict(obj=msg) for msg in obj.toasts]
    amap["commands"] = [to_dict(uicommand) for uicommand in obj.commands]
    return amap


class Rezulto(Returnable):
    """Well-organized successful response object.

    When your action succeeds you should return a Rezulto.
    Unsuccessful operations raise MalbonaRezulto instead.
    """

    level = "success"
    status_int = 200  # HTTP response code indicating success


class MalbonaRezulto(Returnable, Exception):
    """Base class for exceptions raised by actions."""

    level = "danger"
    status_int = 400  # HTTP response code indicating invalid request

    def __init__(
        self,
        status_int: int = 400,
        title: str = "",
        plain: str = "",
        html: str = "",
        level: str = "danger",
        invalid: Optional[DictStr] = None,
        **kw,
    ):  # noqa
        Returnable.__init__(self, **kw)
        self.status_int = status_int
        self.invalid = invalid or {}
        if title or plain or html:
            self.add_toast(title=title, level=level, plain=plain, html=html)


@to_dict.register(obj=MalbonaRezulto, flavor="")
def malbona_to_dict(obj, flavor="", **kw):
    """Convert a MalbonaRezulto to a dictionary."""
    amap = returnable_to_dict(obj=obj, flavor="", **kw)
    amap["invalid"] = obj.invalid
    return amap



#!/usr/bin/env python3

"""Installer for kerno."""

from sys import version_info

# http://peak.telecommunity.com/DevCenter/setuptools#developer-s-guide
# from distutils.core import setup
from setuptools import setup, find_packages

if version_info[:2] < (3, 6):
    raise RuntimeError("Kerno requires Python 3.6+")

with open("README.rst", encoding="utf-8") as f:
    long_description = f.read()

setup(
    url="https://github.com/nandoflorestan/kerno",
    name="kerno",
    author="Nando Florestan",
    version="0.5.0",
    license="MIT",
    # https://setuptools.readthedocs.io/en/latest/setuptools.html#using-find-packages
    packages=find_packages(exclude=["tests.*", "tests"]),
    include_package_data=True,
    author_email="nandoflorestan@gmail.com",
    description="Framework for the application service layer, "
    "approximating Uncle Bob Martin's Clean Architecture.",
    long_description=long_description,
    zip_safe=False,
    test_suite="tests",
    install_requires=["bag>=3.0", "reg==0.12"],
    keywords=[
        "pyramid",
        "sqlalchemy",
        "service layer",
        "action",
        "Clean Architecture",
    ],
    classifiers=[  # https://pypi.org/pypi?:action=list_classifiers
        # "Development Status :: 3 - Alpha",
        "Development Status :: 4 - Beta",
        # "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Framework :: Pyramid",
        "Topic :: Database",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Internet :: WWW/HTTP :: WSGI",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Application Frameworks",
    ],
)



"""Variables I commonly use when writing Python typing annotations."""

from typing import Any, Dict, TypeVar

# For generic functions. Represents a model instance of any type.
Entity = TypeVar("Entity")

# These are very common Dict type annotations, I think an alias can be useful
DictStr = Dict[str, Any]
DictInt = Dict[int, Any]



"""Extensible scheme for an app to send out email messages.

We provide only a well-structured API to build messages, such that
classes representing individual email messages are easy to change
and easy to test. And any backend capable of actually sending out
emails can be plugged at the end.
"""

from abc import ABCMeta, abstractmethod
from typing import List, Optional, Tuple, Union

from bag.email_validator import EmailValidator
from bag.reify import reify
from bs4 import BeautifulSoup

from kerno.typing import DictStr


class EmailAddress:
    """Represents an email address, optionally with the person's name."""

    email_validator = EmailValidator()

    def __init__(self, email: str, name: str = ''):
        """Validate *email* and instantiate."""
        self.email_validator.validate_or_raise(email)
        self.email = email
        self.name = name

    def to_mailer(self) -> Union[str, Tuple[str, str]]:
        """Return either email or a 2-tuple (name, email)."""
        return (self.name, self.email) if self.name else self.email

    def __str__(self) -> str:
        return f'"{self.name}" <{self.email}>' if self.name else self.email


class Envelope:
    """Represents the envelope of an email message."""

    def __init__(  # noqa
        self, recipients: List[EmailAddress],
        cc: Optional[List[EmailAddress]] = None,
        bcc: Optional[List[EmailAddress]] = None,
        reply_to: Optional[EmailAddress] = None,
        sender: Optional[EmailAddress] = None,
    ):
        assert recipients
        self.recipients = recipients
        self.cc = cc or []
        self.bcc = bcc or []
        self.reply_to = reply_to
        self.sender = sender

    def to_dict(self) -> DictStr:
        """Return a dict with Python primitive types within."""
        return {
            'recipients': [str(r) for r in self.recipients],
            'cc': [str(r) for r in self.cc],
            'bcc': [str(r) for r in self.bcc],
            'reply_to': str(self.reply_to) if self.reply_to else None,
            'sender': str(self.sender) if self.sender else None,
        }


class EmailMessageBase(metaclass=ABCMeta):
    """Abstract base class to build an email message from templates.

    The plain text version is automatically built from the HTML version.

    Subclasses should declare certain static variables::

        class ACertainEmailMessage(EmailMessageBase):

            SUBJECT = 'Life or death matter in {app_name}!!!1'
            HTML_TEMPLATE = 'path/to/template.jinja2'
    """

    def __init__(self, adict: DictStr, envelope: Envelope):
        """:param adict: dictionary for use in templates."""
        self.adict = adict
        self.envelope = envelope

    @reify
    def subject(self) -> str:
        """May be overridden in subclasses to decorate the subject line."""
        return self.SUBJECT.format(**self.adict)  # type: ignore

    @reify
    @abstractmethod
    def html(self) -> str:
        """Must be overridden in subclasses to return the HTML version.

        Usually based on HTML_TEMPLATE. Example implementation::

            @reify
            def rich(self):
                return jinja2.get_template(
                    self.HTML_TEMPLATE).render(self.adict)
        """
        raise NotImplementedError()

    @reify
    def plain(self) -> str:
        """Autogenerate the plain text version from the rich version."""
        return BeautifulSoup(self.html, 'html.parser').get_text()

    def to_dict(self) -> DictStr:
        """Return a dict containing the computed parts of this message."""
        assert self.subject
        assert self.plain or self.html
        return {
            'envelope': self.envelope.to_dict(),
            'subject': self.subject,
            'html': self.html,
            'plain': self.plain,
        }
